/* A Global Alignment Program (GAP4) for Comparing Two Sequences

Developed by Xiaoqiu Huang based on discussion with Doug Brutlag

Copyright (c) 2005 Xiaoqiu Huang
All Rights Reserved.  E-mail: xqhuang@cs.iastate.edu

Permission to use, copy, modify, and distribute this software and its
documentation for educational, research and non-profit purposes, without
fee, and without a written agreement is hereby granted, provided that the
above copyright notice, this paragraph and the following three paragraphs
appear in all copies.

Permission to incorporate this software into commercial products may be
obtained from Iowa State University Research Foundation, Inc.

	      Xiaoqiu Huang
	      Department of Computer Science
	      Iowa State University
	      Ames, IA 50011
	      E-mail: xqhuang@cs.iastate.edu
	      WWW: http://www.cs.iastate.edu/~xqhuang

    Proper attribution of the author as the source of the software would
    be appreciated:

    The GAP4 program computes an optimal global alignemnt of two sequences.
    An optimal alignment is one with the maximum score among all alignments
    under all assignments. It delivers the alignment in linear space,
    so long sequences can be aligned. 

    Users supply a file of parameter sets of the form:

# Comment line. A complete set of parameters are given in order per line.
# Output letter  Matrix file name  Gap open  Gap extension  Block  Transition
l  BS45  10  2  40  20
m  BS62  12  3  40  20
h  BS100  14  4  40  20

    One set of parameters is given on a line.
    Column 1 is an output letter for the parameter set.
    Column 2 is the name of a substitution matrix file.
    Column 3 is a gap open penalty.
    Column 4 is a gap extension penalty.
    Column 5 is a difference block penalty.
    Column 6 is a transition penalty from one set of parameters to another.

    In each substitution matrix file,
    users define an alphabet of characters appearing
    in the sequences and a matrix that gives the substitution score
    for each pair of symbols in the alphabet. The 127 ASCII characters
    are eligible. The file may begin with comment lines, followed by
    a line listing the characters in the alphabet and the lower
    triangle of the matrix. An example file is given below:

#  Matrix made by matblas from blosum62_3.iij
#  * column uses minimum score
#  BLOSUM Clustered Scoring Matrix in 1/3 Bit Units
#  Blocks Database = /data/blocks_5.0/blocks.dat
#  Cluster Percentage: >= 62
#  Entropy =   0.6979, Expected =  -0.5209
ARNDCQEGHILKMFPSTWYVBZX
 6
-2  8
-2 -1  8
-3 -2  2  9
-1 -5 -4 -5 13
-1  1  0  0 -4  8
-1  0  0  2 -5  3  7
 0 -3 -1 -2 -4 -3 -3  8
-2  0  1 -2 -4  1  0 -3 11
-2 -4 -5 -5 -2 -4 -5 -6 -5  6
-2 -3 -5 -5 -2 -3 -4 -5 -4  2  6
-1  3  0 -1 -5  2  1 -2 -1 -4 -4  7
-1 -2 -3 -5 -2 -1 -3 -4 -2  2  3 -2  8
-3 -4 -4 -5 -4 -5 -5 -5 -2  0  1 -5  0  9
-1 -3 -3 -2 -4 -2 -2 -3 -3 -4 -4 -2 -4 -5 11
 2 -1  1  0 -1  0  0  0 -1 -4 -4  0 -2 -4 -1  6
 0 -2  0 -2 -1 -1 -1 -2 -3 -1 -2 -1 -1 -3 -2  2  7
-4 -4 -6 -6 -3 -3 -4 -4 -4 -4 -2 -4 -2  1 -5 -4 -4 16
-3 -3 -3 -5 -4 -2 -3 -5  3 -2 -2 -3 -1  4 -4 -3 -2  3 10
 0 -4 -4 -5 -1 -3 -4 -5 -5  4  1 -3  1 -1 -4 -2  0 -4 -2  6
-2 -2  5  6 -5  0  1 -1 -1 -5 -5 -1 -4 -5 -3  0 -1 -6 -4 -5  5
-1  0  0  1 -5  5  6 -3  0 -5 -4  1 -2 -5 -2  0 -1 -4 -3 -4  0  5
-1 -2 -2 -2 -3 -1 -1 -2 -2 -2 -2 -1 -1 -2 -2 -1 -1 -3 -2 -1 -2 -1 -2

    The GAP4 program is written in C and runs under Unix and Linux systems.
    We think that the program is portable to many machines.

    Sequences to be analyzed are stored in separate files.
    Sequences must be in FASTA format. The first line begins with the symbol '>'
    followed by the name of the sequence. The sequence is on the remaining lines.
    Protein sequences must be in upper case.
    DNA sequences could be in upper or lower case.
    A sample sequence file is shown below.

>Human-beta
VHLTPEEKSAVTALWGKVNVDEVGGEALGRLLVVYPWTQRFFESFGDLSTPDAVMGNPKV
KAHGKKVLGAFSDGLAHLDNLKGTFATLSELHCDKLHVDPENFRLLGNVLVCVLAHHFGK
EFTPPVQAAYQKVVAGVANALAHKYH

    To find an optimal alignment of two sequences in files A and B,
    use a command of form

	   gap4  A  B  pf  > result

    where pf is the name of a parameter file and result
    is the name of an output file.

*/

#include   <stdio.h>

#define  LINELEN    60   /* lenght of an output line */
#define  LineLen   900   /* max length of a line in the parameter file */
static int numrows = 30; /* max number of rows in the basis */

static int match, mismh;		/* max and min substitution weights */
static int isdna;                       /* 1, DNA; 0, protein */
static int  neginf = -100000;	/* negative infinity */

int  nom = 1;	/* number of substitution tables */
typedef struct MATPEN
  { int v[128][128];	/* substitution table */
    int q, r, qr;	/* gap penalties */
    int pay, w;		/* block penalty and transition penalty */
    char let;		/* output letter for each alignment column */
    char sname[LineLen + 2];	/* name of a substitution matrix */
  }  matpentype,  *matpenptr;	/* for all sets of tables and penalties */
matpenptr  matp;	/* all sets of substitution tables and penalties */
static long  totalmemreq = 0;
static int   maxM, maxN;	/* maximum sequence lengths */

GetParameters(pfname, M, N)
char  *pfname;	/* parameter file name */
int   M, N;	/* max seq lengths */
{ int  symbol;				/* The next character	      */
  int  ms;				/* User-supplied weights      */
  FILE *Sp, *Pfp, *ckopen();
  void *ckalloc();			/* space-allocating function  */
  int   i, j, k;
  char  alph[LineLen + 2], *s;		/* alphabet */
  int   size;				/* size of alphabet */
  char  *name;				/* temporary pointer */
  char  rdline[LineLen + 2];		/* a line */
  char  sfname[LineLen + 2];		/* name of a substitution file */
  char  cline[LineLen + 2];		/* a line of a substitution file */
  char  oflag;		/* flag for a parameter set */
  int   q, r;		/* gap penalties */
  int   pay;		/* constant-cost for gaps */
  int   trw;		/* transition penalty */
  int   tmp;

	isdna = 0;
	Pfp = ckopen(pfname, "r");
	for ( k = 0; fgets(rdline, LineLen, Pfp) != NULL; )
	 { sscanf(rdline, "%s", sfname);
	   if ( sfname[0] == '#' ) continue;
	   k++;
	 }
	nom = k;
	rewind(Pfp);
	matp = ( matpenptr ) ckalloc( (nom+2) * sizeof(matpentype) );
	for ( k = 0; fgets(rdline, LineLen, Pfp) != NULL; )
	 { sscanf(rdline, "%s", sfname);
	   if ( sfname[0] == '#' ) continue;
	   if ( sscanf(rdline, "%c %s %d %d %d %d",
	        &oflag, sfname, &q, &r, &pay, &trw) < 6 )
	      fatal("A line of wrong format in the parameter file");
	   if ( k >= nom )  fatal("Extra lines in the parameter file");
	   if ( pay < 1 )
	     fatal("Penalty for a major difference is a positive integer");
	   if ( q < 0 )
	     fatal("The gap-open penalty is a nonnegative integer");
	   if ( r < 0 )
	     fatal("The gap-extend penalty is a nonnegative integer");
	   matp[k].let = oflag;
	   matp[k].q = q;
	   matp[k].r = r;
	   matp[k].qr = q + r;
	   matp[k].pay = pay;
	   matp[k].w = trw;
	   strcpy(matp[k].sname, sfname);

	  { /* read a file containing alphabet and substitution weights */
	    Sp = ckopen(sfname, "r");
	    while ( fgets(cline, LineLen, Sp) != NULL )
	     { sscanf(cline, "%s", alph);
	       if ( alph[0] != '#' )  break;
	     }
	    size = strlen(alph);
	    for ( i = 0; i < 128 ; i++ )
	      for ( j = 0; j < 128 ; j++ )
	            matp[k].v[i][j] = 0;
	    match = mismh = 0;
	    for ( i = 0; i < size ; i++ )
	     for ( j = 0; j <= i ; j++ )
		{ (void) fscanf(Sp, "%d", &ms);
		  matp[k].v[alph[i]][alph[j]] = matp[k].v[alph[j]][alph[i]] = ms;
		  if ( ms > match ) match = ms;
		  if ( ms < mismh ) mismh = ms;
		}
	    if ( isdna )
	     { matp[k].v['N']['N'] = mismh;
               matp[k].v['n']['n'] = mismh;
	       matp[k].v['a']['a'] = tmp = matp[k].v['A']['A'];
               matp[k].v['A']['a'] = matp[k].v['a']['A'] = tmp;
	       matp[k].v['c']['c'] = tmp = matp[k].v['C']['C'];
               matp[k].v['C']['c'] = matp[k].v['c']['C'] = tmp;
	       matp[k].v['g']['g'] = tmp = matp[k].v['G']['G'];
               matp[k].v['G']['g'] = matp[k].v['g']['G'] = tmp;
	       matp[k].v['t']['t'] = tmp = matp[k].v['T']['T'];
               matp[k].v['T']['t'] = matp[k].v['t']['T'] = tmp;
	     }
	    fclose(Sp);
	  }
	   k++;
	 }
	fclose(Pfp);
	AllocateSpace(M, N);
	maxM = M;
	maxN = N;
}

typedef struct VECSCA
  { int *CC, *DD, *HH;	/* forward matrices */
    int *RR, *SS, *XX;	/* reverse matrices */
    int c, e, d, h;	/* scalars for four matrices */
    int s, t, g;	/* scalars */
    int mat, len;	/* length and identity number for the parameter set */
  }  vecscatype,  *vecscaptr;	/* for all sets of matrices and scalars */
static vecscaptr  vecs;	/* all sets of matrices and scalars */
static int  *ZF;	/* overall matrix */
static int  *ZR;	/* overall matrix */

typedef struct INITV
  { int c, d, h;	/* initial values */
  }  initvtype,  *initvptr;	/* for all sets of initial values */

typedef struct OPDBPF
  { int   op;		/* operation code */
    int   pk;		/* parameter set index */
    short md;		/* difference block flag */
  }  opdbpftype,  *opdbpfptr;	/* for all sets of initial values */
static opdbpfptr  soplist;	/* list for op, pflag, and diff block */

struct TraceType
   { char c, d, e, h; }; /* a traceback record */
static struct TraceRec
   { struct TraceType  *ten;
                short  rank; } **tracem; /* traceback information */

/* The following definitions are for function diff() */

static int  diff4();
static opdbpfptr  sapp;				/* Current script append ptr */
static int   last;				/* Last script op appended */
static short premd;				/* last major difference flag */
static int   prepk;				/* last parameter set index */

static int no_mat; 				/* number of matches */ 
static int no_mis; 				/* number of mismatches */ 
static int al_len; 				/* length of alignment */
static int no_regs; 				/* number of similar regions */
static initvptr finit, rinit;	/* pointers to initial value records */
static int sti, stj, edi, edj;  /* start and end points of an alignment */
static int mscore;              /* max score of alignments */

						/* Append "Delete k" op */
#define DEL(k, db, pi)                  \
{ if ( ! db )                           \
   { al_len += (k);                     \
     vecs[pi].len += (k);               \
   }                                    \
  if ( last < 0 && premd == db && prepk == pi )  \
    last = sapp[-1].op -= (k);          \
  else                                  \
   { last = sapp->op = -(k);            \
     premd = sapp->md = db;             \
     prepk = sapp->pk = pi;             \
     sapp++;                            \
   }                                    \
}
						/* Append "Insert k" op */
#define INS(k, db, pi)			\
{ if ( ! db )                           \
   { al_len += (k);                     \
     vecs[pi].len += (k);               \
   }                                    \
  if ( last > 0 && premd == db && prepk == pi )  \
    last = sapp[-1].op += (k);          \
  else					\
   { last = sapp->op = (k);             \
     premd = sapp->md = db;             \
     prepk = sapp->pk = pi;             \
     sapp++;                            \
   }                                    \
}
						/* Append "Replace" op */
#define REP(pi) 			\
{ if ( premd ) no_regs++;               \
  last = sapp->op = 0;		        \
  premd = sapp->md = 0;                 \
  prepk = sapp->pk = pi;                \
  sapp++;                               \
  al_len++;                             \
  vecs[pi].len++;                       \
}

static initvptr finit, rinit;	/* pointers to initial value records */

AllocateSpace(M,N) int M,N;
{ int  score;   		/* the max score in LIST */
  int  i, j, k;			/* row and column indices */
  void *ckalloc();		/* space-allocating function */
  int  tensize;
  matpenptr  mpp;		/* pointer to a record in matp */
	
	    /* allocate space for all vectors */
            i = (nom + 1) * sizeof(vecscatype);
	    vecs = ( vecscaptr ) ckalloc(i);
	    j = (N + 1) * sizeof(int);
	    ZF = ( int * ) ckalloc(j);
	    ZR = ( int * ) ckalloc(j);
	    for ( k = 0; k < nom; k++ )
	     { vecs[k].CC = ( int * ) ckalloc(j);
	       vecs[k].DD = ( int * ) ckalloc(j);
	       vecs[k].RR = ( int * ) ckalloc(j);
	       vecs[k].SS = ( int * ) ckalloc(j);
	       vecs[k].HH = ( int * ) ckalloc(j);
	       vecs[k].XX = ( int * ) ckalloc(j);
	       vecs[k].mat = vecs[k].len = 0;
	     }
	    i = (M + N + 2) * sizeof(opdbpftype);
	    soplist = ( opdbpfptr ) ckalloc(i);
	    i = (numrows + 1) * sizeof(struct TraceRec *);
	    tracem = ( struct TraceRec ** ) ckalloc(i);
	    j = (N + 1) * sizeof(struct TraceRec);
	    for ( i = 0; i <= numrows; i++ )
	      tracem[i] = ( struct TraceRec * ) ckalloc(j);
	    tensize = (nom + 1) * sizeof(struct TraceType);
	    for ( i = 0; i <= numrows; i++ )
	      for ( j = 0; j <= N; j++ )
	        tracem[i][j].ten = ( struct TraceType * ) ckalloc(tensize);

            finit = ( initvptr ) ckalloc( (nom + 1) * sizeof ( initvtype ) );
            rinit = ( initvptr ) ckalloc( (nom + 1) * sizeof ( initvtype ) );
}

int UseGAP4(A, B, M, N, dhead1, dhead2)
char A[], B[];		/* sequences */
int  M, N;		/* their lengths */
char dhead1[], dhead2[];	/* their names */
{ int  score;   		/* the max score in LIST */
  int  i, j, k;			/* row and column indices */
  int  tensize;
  matpenptr  mpp;		/* pointer to a record in matp */

	for ( k = 0; k < nom; k++ )
	  vecs[k].mat = vecs[k].len = 0;
  	
	if ( M > maxM || N > maxN )
	  fatal("Sequences are too long");
	
	/*
(void) printf("Pflag  Sub. matrix  Gap-open pen.  Gap-ext. pen.  Diff. block pen.  Tran. pen.\n");
	    for ( k = 0; k < nom; k++ )
	     {
(void) printf(" %c      %10s     %d             %d              %d                %d\n",
                matp[k].let, matp[k].sname, matp[k].q, matp[k].r, matp[k].pay, matp[k].w);
	     }
	    (void) printf("\nUpper Sequence: %s\n", dhead1);
	    (void) printf("      Length: %d\n", M);
	    (void) printf("Lower Sequence: %s\n", dhead2);
	    (void) printf("      Length: %d\n\n", N);  */
            for ( mpp = matp, k = 0; k < nom; k++, mpp++ )
	     { finit[k].c = 0 /* - mpp->w */;
	       finit[k].h = - mpp->pay /* - mpp->w */;
	       finit[k].d = - mpp->q /* - mpp->w */;
	       rinit[k].c = - mpp->w;
	       rinit[k].h = - mpp->pay - mpp->w;
	       rinit[k].d = - mpp->q - mpp->w;
	     }
	    ScoreOnly4(A, B, M, N, 0, finit, 0, rinit);
	    return mscore;
	    /*
	    mscore = sti = stj = edi = edj = 0;
	    FindEndPoints4(A, B, M, N, 0, finit, 0, rinit);
            sapp = soplist;
            last = 0;
	    premd = 1;
	    prepk = 0;
            al_len = no_mat = no_mis = no_regs = 0;
	    score = diff4(A+sti, B+stj, edi-sti, edj-stj, 0, finit, 0, rinit);
(void) printf("Best alignment with dynamic use of parameter sets\n");
(void) printf("Similarity score: %d\n",score);
(void) printf("Percent identity: %d%%\n",
 (al_len == 0 ? 0 : ( (100*no_mat)/al_len ) ) );
(void) printf("Number of matches: %d\n", no_mat);
(void) printf("Number of mismatches: %d\n", no_mis);
(void) printf("Length of gaps: %d\n", al_len-no_mat-no_mis);
(void) printf("Number of similarity blocks: %d\n", no_regs);
(void) printf("Length of similarity blocks: %d\n\n", al_len);
if ( mscore != score ) fatal("Different scores between score-only and alignment");
(void) printf("Begins at (%d, %d) and Ends at (%d, %d)\n",
                                    sti + 1, stj + 1, edi, edj);
	    for ( k = 0; k < nom; k++ )
	     {
(void) printf("For similar regions of type %c, Length: %d;  Percent identity: %d%%\n",
matp[k].let, vecs[k].len, (vecs[k].len == 0 ? 0 : ((100 * vecs[k].mat) / vecs[k].len) ) );
	     }
            display4(A+sti, B+stj, edi-sti, edj-stj, soplist, sti + 1 , stj + 1);
	    return score;  */
}

int GetGAP4Prop(A, B, M, N, dhead1, dhead2, prop)
char A[], B[];		/* sequences */
int  M, N;		/* their lengths */
char dhead1[], dhead2[];	/* their names */
double  *prop;		/* proportions for each parameter set */
{ int  score;   		/* the max score in LIST */
  int  i, j, k;			/* row and column indices */
  int  tensize;
  matpenptr  mpp;		/* pointer to a record in matp */
  double  total;		/* total length */

	for ( k = 0; k < nom; k++ )
	  vecs[k].mat = vecs[k].len = 0;
  	
	if ( M > maxM || N > maxN )
	  fatal("Sequences are too long");
	
	/*
(void) printf("Pflag  Sub. matrix  Gap-open pen.  Gap-ext. pen.  Diff. block pen.  Tran. pen.\n");
	    for ( k = 0; k < nom; k++ )
	     {
(void) printf(" %c      %10s     %d             %d              %d                %d\n",
                matp[k].let, matp[k].sname, matp[k].q, matp[k].r, matp[k].pay, matp[k].w);
	     }
	    (void) printf("\nUpper Sequence: %s\n", dhead1);
	    (void) printf("      Length: %d\n", M);
	    (void) printf("Lower Sequence: %s\n", dhead2);
	    (void) printf("      Length: %d\n\n", N);  */
            for ( mpp = matp, k = 0; k < nom; k++, mpp++ )
	     { finit[k].c = 0 /* - mpp->w */;
	       finit[k].h = - mpp->pay /* - mpp->w */;
	       finit[k].d = - mpp->q /* - mpp->w */;
	       rinit[k].c = - mpp->w;
	       rinit[k].h = - mpp->pay - mpp->w;
	       rinit[k].d = - mpp->q - mpp->w;
	     }
	    
	    mscore = sti = stj = edi = edj = 0;
	    FindEndPoints4(A, B, M, N, 0, finit, 0, rinit);
            sapp = soplist;
            last = 0;
	    premd = 1;
	    prepk = 0;
            al_len = no_mat = no_mis = no_regs = 0;
	    score = diff4(A+sti, B+stj, edi-sti, edj-stj, 0, finit, 0, rinit);
/*
(void) printf("Best alignment with dynamic use of parameter sets\n");
(void) printf("Similarity score: %d\n",score);
(void) printf("Percent identity: %d%%\n",
 (al_len == 0 ? 0 : ( (100*no_mat)/al_len ) ) );
(void) printf("Number of matches: %d\n", no_mat);
(void) printf("Number of mismatches: %d\n", no_mis);
(void) printf("Length of gaps: %d\n", al_len-no_mat-no_mis);
(void) printf("Number of similarity blocks: %d\n", no_regs);
(void) printf("Length of similarity blocks: %d\n\n", al_len);
*/
if ( mscore != score ) fatal("Different scores between score-only and alignment");
/*
(void) printf("Begins at (%d, %d) and Ends at (%d, %d)\n",
                                    sti + 1, stj + 1, edi, edj);  */
            
            total = 0.0;
	    for ( k = 0; k < nom; k++ )
	       total += vecs[k].len;
	    for ( k = 0; k < nom; k++ )
               prop[k] = (double) vecs[k].len / total;
/*
            display4(A+sti, B+stj, edi-sti, edj-stj, soplist, sti + 1 , stj + 1);
*/
	    return score;
}

/* Compute only the score of a best local alignment. */
ScoreOnly4(A, B, M, N, z00, iv00, zmn, ivmn)
char *A, *B; int M, N;	/* two sequences and their lengths */
int  z00, zmn;		/* initial values for the overall matrix ZF */
initvptr iv00, ivmn;	/* pointers to arrays of initial values for matrices */
{ vecscaptr vsp;	  /* pointer to an array of matrices and scalars */
  matpenptr mpp;	  /* pointer to an array of sub tables and penalties */
  int   i, j, k;
  int   c, e, d, h, z;
  int   t, zp, zi, zj, ze, zd;
  int   g, x, y, tmp;
  int  ai, bj;		  /* integers for A[i] and B[j] */
  int  maxs;		  /* max score */

  maxs = 0;	/* Forward phase */
  ZF[0] = z00;
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->CC[0] = c = (tmp = iv00[k].c) > z00 ? tmp : z00;
     vsp->e = e = c - mpp->q;
     vsp->g = (h = iv00[k].h) > (y = c - mpp->pay) ? h : y;
     vsp->t = (d = iv00[k].d) > e ? d : e;
   }
  for (j = 1; j <= N; j++)
   { z = neginf;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r;
        vsp->CC[j] = c = e > (g = vsp->g) ? e : g;
	if ( z < c ) z = c;
        vsp->DD[j] = neginf;
        vsp->HH[j] = g;
      }
     if ( z < 0 ) z = 0;
     ZF[j] = z;
   }
  		
  for (i = 1; i <= M; i++)
    { z = neginf;
      for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
       { vsp->s = vsp->CC[0];
         vsp->t = t = vsp->t - mpp->r;
         vsp->CC[0] = vsp->c = c = t > (g = vsp->g) ? t : g;
	 if ( z < c ) z = c;
	 vsp->e = neginf;
	 vsp->h = g;
       }
      zp = ZF[0];
      if ( z < 0 ) z = 0;
      ZF[0] = z;
      ai = A[i];
	  	 
      for (j = 1; j <= N; j++)
        { zi = z; z = neginf;
	  zj = ZF[j];
	  bj = B[j];
          for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	   { if ( (ze = zi - mpp->w) < (c = vsp->c) ) ze = c;
	     vsp->e = e = (d = ze - mpp->qr) > (y = vsp->e - mpp->r) ? d : y;
	     if ( (zd = zj - mpp->w) < (x = vsp->CC[j]) ) zd = x;
	     if ( ze < zd ) ze = zd;
	     if ( (ze -= mpp->pay) < (y = vsp->HH[j]) ) ze = y;
	     if ( (h = vsp->h) < ze ) h = ze;
	     if ( (tmp = zd - mpp->qr) > (d = vsp->DD[j] - mpp->r) ) d = tmp;
	     if ( (y = zp - mpp->w) < 0 ) y = 0;  /* start over */
	     if ( (tmp = vsp->s) < y ) tmp = y;
	     c = tmp + mpp->v[ai][bj];
             if ( c < d ) c = d;
             if ( c < e ) c = e;
	     if ( c < h ) c = h;
	     vsp->s = x;
	     vsp->CC[j] = vsp->c = c;
	     vsp->DD[j] = d;
	     vsp->HH[j] = vsp->h = h;
	     if ( z < c ) z = c;
	   }
          if ( z < 0 ) z = 0;
	  if ( maxs < z )
	   { maxs = z; edi = i; edj = j; }
	  zp = zj;
	  ZF[j] = z;
        }
    }

  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->DD[0] = vsp->t;
     vsp->HH[0] = vsp->g;
   }
  mscore = maxs;
}

/* Find the score and end points of a best local alignment
and save them in mscore, (sti, stj), and (edi, edj). */
FindEndPoints4(A, B, M, N, z00, iv00, zmn, ivmn)
char *A, *B; int M, N;	/* two sequences and their lengths */
int  z00, zmn;		/* initial values for the overall matrix ZF */
initvptr iv00, ivmn;	/* pointers to arrays of initial values for matrices */
{ vecscaptr vsp;	  /* pointer to an array of matrices and scalars */
  matpenptr mpp;	  /* pointer to an array of sub tables and penalties */
  int   i, j, k;
  int   c, e, d, h, z;
  int   t, zp, zi, zj, ze, zd;
  int   g, x, y, tmp;
  int  ai, bj;		  /* integers for A[i] and B[j] */
  int  maxs;		  /* max score */

  maxs = 0;	/* Forward phase */
  ZF[0] = z00;
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->CC[0] = c = (tmp = iv00[k].c) > z00 ? tmp : z00;
     vsp->e = e = c - mpp->q;
     vsp->g = (h = iv00[k].h) > (y = c - mpp->pay) ? h : y;
     vsp->t = (d = iv00[k].d) > e ? d : e;
   }
  for (j = 1; j <= N; j++)
   { z = neginf;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r;
        vsp->CC[j] = c = e > (g = vsp->g) ? e : g;
	if ( z < c ) z = c;
        vsp->DD[j] = neginf;
        vsp->HH[j] = g;
      }
     if ( z < 0 ) z = 0;
     ZF[j] = z;
   }
  for (i = 1; i <= M; i++)
    { z = neginf;
      for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
       { vsp->s = vsp->CC[0];
         vsp->t = t = vsp->t - mpp->r;
         vsp->CC[0] = vsp->c = c = t > (g = vsp->g) ? t : g;
	 if ( z < c ) z = c;
	 vsp->e = neginf;
	 vsp->h = g;
       }
      zp = ZF[0];
      if ( z < 0 ) z = 0;
      ZF[0] = z;
      ai = A[i];
      for (j = 1; j <= N; j++)
        { zi = z; z = neginf;
	  zj = ZF[j];
	  bj = B[j];
          for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	   { if ( (ze = zi - mpp->w) < (c = vsp->c) ) ze = c;
	     vsp->e = e = (d = ze - mpp->qr) > (y = vsp->e - mpp->r) ? d : y;
	     if ( (zd = zj - mpp->w) < (x = vsp->CC[j]) ) zd = x;
	     if ( ze < zd ) ze = zd;
	     if ( (ze -= mpp->pay) < (y = vsp->HH[j]) ) ze = y;
	     if ( (h = vsp->h) < ze ) h = ze;
	     if ( (tmp = zd - mpp->qr) > (d = vsp->DD[j] - mpp->r) ) d = tmp;
	     if ( (y = zp - mpp->w) < 0 ) y = 0;  /* start over */
	     if ( (tmp = vsp->s) < y ) tmp = y;
	     c = tmp + mpp->v[ai][bj];
             if ( c < d ) c = d;
             if ( c < e ) c = e;
	     if ( c < h ) c = h;
	     vsp->s = x;
	     vsp->CC[j] = vsp->c = c;
	     vsp->DD[j] = d;
	     vsp->HH[j] = vsp->h = h;
	     if ( z < c ) z = c;
	   }
          if ( z < 0 ) z = 0;
	  if ( maxs < z )
	   { maxs = z; edi = i; edj = j; }
	  zp = zj;
	  ZF[j] = z;
        }
    }
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->DD[0] = vsp->t;
     vsp->HH[0] = vsp->g;
   }
  if ( (mscore = maxs) == 0 ) return;
  M = edi; N = edj;

  maxs = 0;	/* Reverse phase */
  ZR[N] = zmn;
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->RR[N] = c = (tmp = ivmn[k].c) > zmn ? tmp : zmn;
     vsp->e = e = c - mpp->q;
     vsp->g = (h = ivmn[k].h) > (y = c - mpp->pay) ? h : y;
     vsp->t = (d = ivmn[k].d) > e ? d : e;
   }
  for (j = N-1; j >= 0; j--)
   { z = neginf;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r;
        vsp->RR[j] = c = e > (g = vsp->g) ? e : g;
	if ( z < c ) z = c;
        vsp->SS[j] = neginf;
        vsp->XX[j] = g;
      }
     ZR[j] = z;
   }
  for (i = M-1; i >= 0; i--)
    { z = neginf;
      for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
       { vsp->s = vsp->RR[N];
         vsp->t = t = vsp->t - mpp->r;
         vsp->RR[N] = vsp->c = c = t > (g = vsp->g) ? t : g;
	 if ( z < c ) z = c;
	 vsp->e = neginf;
	 vsp->h = g;
       }
      zp = ZR[N];
      ZR[N] = z;
      ai = A[i+1];
      for (j = N-1; j >= 0; j--)
        { zi = z; z = neginf;
	  zj = ZR[j];
	  bj = B[j+1];
          for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	   { if ( (ze = zi - mpp->w) < (c = vsp->c) ) ze = c;
	     vsp->e = e = (d = ze - mpp->qr) > (y = vsp->e - mpp->r) ? d : y;
	     if ( (zd = zj - mpp->w) < (x = vsp->RR[j]) ) zd = x;
	     if ( ze < zd ) ze = zd;
	     if ( (ze -= mpp->pay) < (y = vsp->XX[j]) ) ze = y;
	     if ( (h = vsp->h) < ze ) h = ze;
	     if ( (tmp = zd - mpp->qr) > (d = vsp->SS[j] - mpp->r) ) d = tmp;
	     if ( (tmp = vsp->s) < (y = zp - mpp->w) ) tmp = y;
	     c = tmp + mpp->v[ai][bj];
             if ( c < d ) c = d;
             if ( c < e ) c = e;
	     if ( c < h ) c = h;
	     vsp->s = x;
	     vsp->RR[j] = vsp->c = c;
	     vsp->SS[j] = d;
	     vsp->XX[j] = vsp->h = h;
	     if ( z < c ) z = c;
	   }
	  if ( maxs < z )
	    { maxs = z; sti = i; stj = j; }
	  zp = zj;
	  ZR[j] = z;
        }
    }
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->SS[N] = vsp->t;
     vsp->XX[N] = vsp->g;
   }
  if ( mscore != maxs )
    fatal("Different scores between forward and reverse passes");
}

/* diff4(A,B,M,N,z00,iv00,zmn,ivmn) returns the score of an optimum
   conversion between A[1..M] and B[1..N] such that the upper left and
   lower right corners are initialized to the given values. */

int diff4(A, B, M, N, z00, iv00, zmn, ivmn)
char *A, *B; int M, N;	/* two sequences and their lengths */
int  z00, zmn;		/* initial values for the overall matrix ZF */
initvptr iv00, ivmn;	/* pointers to arrays of initial values for matrices */

{ int  midi, midj, midc;	/* Midpoint and cost */
  int  zmidf, zmidr;		/* C, D & H values at midpoint */
  initvptr formid, revmid;	/* pointers to arrays of new initial values */
  void *ckalloc();		/* space-allocating function */
  vecscaptr vsp;	  /* pointer to an array of matrices and scalars */
  matpenptr mpp;	  /* pointer to an array of sub tables and penalties */
  int   i, j, k;

{ int   c, e, d, h, z;
  int   t, zp, zi, zj, ze, zd, *va;
  int   g, x, y, tmp;
  struct TraceRec  *trow; /* one row of tracem */
  struct TraceType *ten;  /* pointer to one entry of tracem */
  char   letc;		  /* a letter to indicate matrix type */
  char zetr, zdtr; 	  /* traceback characters */
  int  ai, bj;		  /* integers for A[i] and B[j] */

/* Boundary cases: M <= numrows */

if ( M <= numrows )
{ trow = tracem[M]; ten = trow[N].ten;
  ZR[N] = zmn; trow[N].rank = 0;
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { if ( (tmp = ivmn[k].c) > (y = zmn - mpp->w) )
      { vsp->RR[N] = c = tmp; ten->c = 'c'; }
     else
      { vsp->RR[N] = c = y; ten->c = 's'; }
     vsp->e = e = c - mpp->q; ten->e = ten->c;
     if ( (h = ivmn[k].h) > (y = c - mpp->pay) )
      { vsp->g = h; ten->h = 'h'; }
     else
      { vsp->g = y; ten->h = ten->c == 'c' ? 'x' : 'p'; }
     if ( (d = ivmn[k].d) > e )
      { vsp->t = d; ten->d = 'd'; }
     else
      { vsp->t = e; ten->d = ten->c; }
     ten++;
   }
  for (j = N-1; j >= 0; j--)
   { ten = trow[j].ten;
     z = neginf; trow[j].rank = 0;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r; ten->e = 'e';
	if ( e > (g = vsp->g) )
	 { vsp->RR[j] = c = e; ten->c = 'e'; }
	else
	 { vsp->RR[j] = c = g; ten->c = 'h'; }
	if ( z < c ) { z = c; trow[j].rank = k; }
	vsp->SS[j] = neginf; ten->d = 'c';
	vsp->XX[j] = g; ten->h = 'g';
        ten++;
      }
     ZR[j] = z;
   }
  for (i = M-1; i >= 0; i--)
   { trow = tracem[i]; ten = trow[N].ten;
     z = neginf; trow[N].rank = 0;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->s = vsp->RR[N];
        vsp->t = t = vsp->t - mpp->r; ten->d = 'd';
	if ( t > (g = vsp->g) )
	 { vsp->RR[N] = vsp->c = c = t; ten->c = 'd'; }
	else
	 { vsp->RR[N] = vsp->c = c = g; ten->c = 'h'; }
	if ( z < c ) { z = c; trow[N].rank = k; }
	vsp->e = neginf; ten->e = 'c';
	vsp->h = g; ten->h = 'h';
        ten++;
      }
     zp = ZR[N];
     ZR[N] = z;
     ai = A[i+1];
     for (j = N-1; j >= 0; j--)
       { ten = trow[j].ten;
         zi = z; z = neginf; trow[j].rank = 0;
	 zj = ZR[j];
	 bj = B[j+1];
	 for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	  { if ( (ze = zi - mpp->w) < (c = vsp->c) )
	     { ze = c; zetr = 'c'; }
	    else
	      zetr = 's';
	    if ( (d = ze - mpp->qr) > (y = vsp->e - mpp->r) )
	     { vsp->e = e = d; ten->e = zetr; }
	    else
	     { vsp->e = e = y; ten->e = 'e'; }
	    if ( (zd = zj - mpp->w) < (x = vsp->RR[j]) )
	     { zd = x; zdtr = 'c'; }
	    else
	       zdtr = 's';
	    if ( ze < zd ) 
	     { ze = zd; ten->h = zdtr == 'c' ? 'x' : 'p'; }
	    else
	     { ten->h = zetr == 'c' ? 'y' : 'r'; }
	    if ( (ze -= mpp->pay) < (y = vsp->XX[j]) ) { ze = y; ten->h = 'h'; }
	    if ( (h = vsp->h) < ze ) h = ze;
	    else  ten->h = 'g';
	    if ( (tmp = zd - mpp->qr) > (d = vsp->SS[j] - mpp->r) )
	     { d = tmp; ten->d = zdtr; }
	    else
	       ten->d = 'd';
	    if ( (tmp = vsp->s) < (y = zp - mpp->w) )
	     { tmp = y; ten->c = 's'; }
	    else
	       ten->c = 'c';
	    c = tmp + mpp->v[ai][bj];
            if ( c < d ) { c = d; ten->c = 'd'; }
            if ( c < e ) { c = e; ten->c = 'e'; }
	    if ( c < h ) { c = h; ten->c = 'h'; }
	    vsp->s = x;
	    vsp->RR[j] = vsp->c = c;
	    vsp->SS[j] = d;
	    vsp->XX[j] = vsp->h = h;
	    if ( z < c ) { z = c; trow[j].rank = k; }
            ten++;
	  }
         zp = zj;
	 ZR[j] = z;
       }
   }
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->SS[N] = vsp->t;
     vsp->XX[N] = vsp->g;
   }

  midc = ZR[0] + z00; letc = 's'; zp = tracem[0][0].rank;
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { tmp = vsp->XX[0] + iv00[k].h + mpp->pay + mpp->w;
     if ( midc < tmp || ( midc == tmp && premd && prepk == k ) )
      { midc = tmp; letc = 'h'; zp = k; }
     tmp = vsp->RR[0] + iv00[k].c + mpp->w;
     if ( midc < tmp || ( midc == tmp && ! premd && prepk == k ) )
      { midc = tmp; letc = 'c'; zp = k; }
     tmp = vsp->SS[0] + iv00[k].d + mpp->q + mpp->w;
     if ( midc < tmp || ( midc == tmp && ! premd && prepk == k && last < 0 ) )
      { midc = tmp; letc = 'd'; zp = k; }
   }    /* if a tie, use the same parameter set */

  i = j = 0; k = zp;
  while ( i < M || j < N )
   { switch ( letc )
{  case 'c' : letc = tracem[i][j].ten[k].c;
              if ( letc == 'c' || letc == 's' )
                { if ( i >= M || j >= N )
		    fatal("Row/column index exceeds the limit");
		  i++; j++;
		  REP(k)
	          if ( A[i] == B[j] && matp[k].v[A[i]][B[j]] > 0 ||
	                        isdna && matp[k].v[A[i]][B[j]] > 0 )
	           { no_mat++; vecs[k].mat++; }
	          else
	            no_mis++;
		}
	      break;
   case 's' : k = tracem[i][j].rank; letc = 'c';
	      break;
   case 'd' : letc = tracem[i][j].ten[k].d;
              if ( i >= M ) fatal("D: row index exceeds the limit");
	      i++;
	      DEL(1, 0, k)
	      break;
   case 'e' : letc = tracem[i][j].ten[k].e;
              if ( j >= N ) fatal("E: column index exceeds the limit");
	      j++;
	      INS(1, 0, k)
	      break;
   case 'h' : letc = tracem[i][j].ten[k].h;
              if ( letc == 'h' || letc == 'x' || letc == 'p' )
                { if ( i >= M ) fatal("Row index exceeds the limit");
	          i++;
		  DEL(1, 1, k)
		  if ( letc == 'x' ) letc = 'c';
		  else
		    if ( letc == 'p' ) letc = 's';
		}
              else
                { if ( j >= N ) fatal("Column index exceeds the limit");
	          j++;
		  INS(1, 1, k)
		  if ( letc == 'g' ) letc = 'h';
		  else
		    if ( letc == 'y' ) letc = 'c';
		    else
		      if ( letc == 'r' ) letc = 's';
		      else
                        fatal("Wrong traceback letter");
		}
              break;
    default : printf("Letter %c\n", letc);
              fatal("Wrong traceback letter");
}
   }
 return midc;
}

/* Divide: Find optimum midpoint (midi,midj) of cost midc */

  midi = M/2;	/* Forward phase */
  ZF[0] = z00;	/* Compute vecs[*].CC(M/2,*), .DD(M/2,*) & .HH(M/2,*) */
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->CC[0] = c = (tmp = iv00[k].c) > (y = z00 - mpp->w) ? tmp : y;
     vsp->e = e = c - mpp->q;
     vsp->g = (h = iv00[k].h) > (y = c - mpp->pay) ? h : y;
     vsp->t = (d = iv00[k].d) > e ? d : e;
   }
  for (j = 1; j <= N; j++)
   { z = neginf;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r;
        vsp->CC[j] = c = e > (g = vsp->g) ? e : g;
	if ( z < c ) z = c;
        vsp->DD[j] = neginf;
        vsp->HH[j] = g;
      }
     ZF[j] = z;
   }
  for (i = 1; i <= midi; i++)
    { z = neginf;
      for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
       { vsp->s = vsp->CC[0];
         vsp->t = t = vsp->t - mpp->r;
         vsp->CC[0] = vsp->c = c = t > (g = vsp->g) ? t : g;
	 if ( z < c ) z = c;
	 vsp->e = neginf;
	 vsp->h = g;
       }
      zp = ZF[0];
      ZF[0] = z;
      ai = A[i];
      for (j = 1; j <= N; j++)
        { zi = z; z = neginf;
	  zj = ZF[j];
	  bj = B[j];
          for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	   { if ( (ze = zi - mpp->w) < (c = vsp->c) ) ze = c;
	     vsp->e = e = (d = ze - mpp->qr) > (y = vsp->e - mpp->r) ? d : y;
	     if ( (zd = zj - mpp->w) < (x = vsp->CC[j]) ) zd = x;
	     if ( ze < zd ) ze = zd;
	     if ( (ze -= mpp->pay) < (y = vsp->HH[j]) ) ze = y;
	     if ( (h = vsp->h) < ze ) h = ze;
	     if ( (tmp = zd - mpp->qr) > (d = vsp->DD[j] - mpp->r) ) d = tmp;
	     if ( (tmp = vsp->s) < (y = zp - mpp->w) ) tmp = y;
	     c = tmp + mpp->v[ai][bj];
             if ( c < d ) c = d;
             if ( c < e ) c = e;
	     if ( c < h ) c = h;
	     vsp->s = x;
	     vsp->CC[j] = vsp->c = c;
	     vsp->DD[j] = d;
	     vsp->HH[j] = vsp->h = h;
	     if ( z < c ) z = c;
	   }
	  zp = zj;
	  ZF[j] = z;
        }
    }
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->DD[0] = vsp->t;
     vsp->HH[0] = vsp->g;
   }

/* Reverse phase */
  ZR[N] = zmn;	/* Compute vecs[*].RR(M/2,*), .SS(M/2,*) & .XX(M/2,*) */
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->RR[N] = c = (tmp = ivmn[k].c) > (y = zmn - mpp->w) ? tmp : y;
     vsp->e = e = c - mpp->q;
     vsp->g = (h = ivmn[k].h) > (y = c - mpp->pay) ? h : y;
     vsp->t = (d = ivmn[k].d) > e ? d : e;
   }
  for (j = N-1; j >= 0; j--)
   { z = neginf;
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { vsp->e = e = vsp->e - mpp->r;
        vsp->RR[j] = c = e > (g = vsp->g) ? e : g;
	if ( z < c ) z = c;
        vsp->SS[j] = neginf;
        vsp->XX[j] = g;
      }
     ZR[j] = z;
   }
  for (i = M-1; i >= midi; i--)
    { z = neginf;
      for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
       { vsp->s = vsp->RR[N];
         vsp->t = t = vsp->t - mpp->r;
         vsp->RR[N] = vsp->c = c = t > (g = vsp->g) ? t : g;
	 if ( z < c ) z = c;
	 vsp->e = neginf;
	 vsp->h = g;
       }
      zp = ZR[N];
      ZR[N] = z;
      ai = A[i+1];
      for (j = N-1; j >= 0; j--)
        { zi = z; z = neginf;
	  zj = ZR[j];
	  bj = B[j+1];
          for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
	   { if ( (ze = zi - mpp->w) < (c = vsp->c) ) ze = c;
	     vsp->e = e = (d = ze - mpp->qr) > (y = vsp->e - mpp->r) ? d : y;
	     if ( (zd = zj - mpp->w) < (x = vsp->RR[j]) ) zd = x;
	     if ( ze < zd ) ze = zd;
	     if ( (ze -= mpp->pay) < (y = vsp->XX[j]) ) ze = y;
	     if ( (h = vsp->h) < ze ) h = ze;
	     if ( (tmp = zd - mpp->qr) > (d = vsp->SS[j] - mpp->r) ) d = tmp;
	     if ( (tmp = vsp->s) < (y = zp - mpp->w) ) tmp = y;
	     c = tmp + mpp->v[ai][bj];
             if ( c < d ) c = d;
             if ( c < e ) c = e;
	     if ( c < h ) c = h;
	     vsp->s = x;
	     vsp->RR[j] = vsp->c = c;
	     vsp->SS[j] = d;
	     vsp->XX[j] = vsp->h = h;
	     if ( z < c ) z = c;
	   }
	  zp = zj;
	  ZR[j] = z;
        }
    }
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { vsp->SS[N] = vsp->t;
     vsp->XX[N] = vsp->g;
   }
   
  midc = ZF[N] + ZR[N];	/* Find optimal midpoint */
  midj = N;
  for ( j = N; j >= 0; j-- )
   { if ( (c = ZF[j] + ZR[j]) > midc )
      { midc = c;
        midj = j;
      }
     for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
      { if ( (c = vsp->HH[j] + vsp->XX[j] + mpp->pay + mpp->w) > midc )
         { midc = c;
           midj = j;
         }
        if ( (c = vsp->DD[j] + vsp->SS[j] + mpp->q + mpp->w) > midc)
         { midc = c;
           midj = j;
         }
        if ( (c = vsp->CC[j] + vsp->RR[j] + mpp->w) > midc )
         { midc = c;
           midj = j;
         }
      }
   }
}

/* Conquer: recursively around midpoint */

  formid = ( initvptr ) ckalloc( (nom + 1) * sizeof ( initvtype ) );
  revmid = ( initvptr ) ckalloc( (nom + 1) * sizeof ( initvtype ) );
  zmidf = ZF[midj];
  zmidr = ZR[midj];
  for ( vsp = vecs, mpp = matp, k = 0; k < nom; k++, vsp++, mpp++ )
   { formid[k].c = vsp->CC[midj];
     formid[k].d = vsp->DD[midj];
     formid[k].h = vsp->HH[midj];
     revmid[k].c = vsp->RR[midj];
     revmid[k].d = vsp->SS[midj];
     revmid[k].h = vsp->XX[midj];
   }
  (void) diff4(A, B, midi, midj, z00, iv00, zmidr, revmid);
  (void) diff4(A+midi, B+midj, M-midi, N-midj, zmidf, formid, zmn, ivmn);
  free(formid);
  free(revmid);
  return midc;
}

/* Alignment display routine */

static char ALINE[LINELEN+1], BLINE[LINELEN+1], CLINE[LINELEN+1];
static char PLINE[LINELEN+1];

display4(A, B, M, N, slist, AP, BP)
char A[], B[];		/* sequences */
int M, N;		/* their lengths */
opdbpfptr slist;        /* list of operations, diff blocks, para. flags */
int AP, BP;		/* start positions in sequences */
{ char *a, *b, *c, *p;
  int   i,  j, op;
  int   lines, ap, bp;
  int gsize;     /* current gap length */
  int gcount;    /* length of a gap removed */
  int flag;      /* 1, print */
  short cmd;	 /* current major difference flag */
  char let;	 /* parameter set flag */
  int  pk;	 /* parameter set index */

  i = j = op = lines = gsize = gcount = 0; let = '$';
  ap = AP;
  bp = BP;
  a = ALINE;
  b = BLINE;
  c = CLINE;
  p = PLINE;
  while (i < M || j < N)
    { if (op == 0 && slist->op == 0)
        { op = slist->op;
	  cmd = slist->md;
	  pk = slist->pk; slist++;
	  let = matp[pk].let;
          *a = A[++i];
          *b = B[++j];
	  *c++ = (*a == *b || isdna && matp[pk].v[*a][*b] > 0 ) ? '|' : ' ';
	  *p++ = let; 
	  a++;
	  b++;
	  gsize = 0;
        }
      else
        { if (op == 0)
           { op = slist->op; gsize = 0; cmd = slist->md;
	     pk = slist->pk; slist++; let = matp[pk].let;
	   }
          if (op > 0)
            { *a++ = ' ';
              *b++ = B[++j];
              op--;
            }
          else
            { *a++ = A[++i];
              *b++ = ' ';
              op++;
            }
	  if ( cmd )
	   { *c++ = '+'; gsize++; }
	  else
             *c++ = '-';
	  *p++ = let; 
        }
      if (a >= ALINE+LINELEN || i >= M && j >= N)
        { *a = *b = *c = *p = '\0';
	  flag = 0;
	  if ( gsize < 2 * LINELEN )
	      flag = 1;
	  else
	   if ( op >= 0 && op < LINELEN || op < 0 && -op < LINELEN )
	    { if ( gcount > 0 )
	       { (void) printf("\n              ~~~ %d bp removed ~~~\n", gcount);
	         gcount = 0;
	       }
	      flag = 1;
	    }
	   else
	     gcount += LINELEN;
	  if ( flag )
	   { (void) printf("\n%7d ",LINELEN*lines);
             for (b = ALINE+10; b <= a; b += 10)
               (void) printf("    .    :");
             if (b <= a+5)
                (void) printf("    .");
             (void) printf("\n%7d %s\n        %s\n%7d %s\n\n        %s\n",
	                              ap, ALINE, CLINE, bp, BLINE, PLINE);
	   }
	  lines++;
	  ap = AP + i;
	  bp = BP + j;
          a = ALINE;
          b = BLINE;
          c = CLINE;
          p = PLINE;
        }
    }
}
